@@ -4,27 +4,29 @@ from decimal import Decimal
 
 from sql.aggregate import Sum
 
-from trytond import backend
 from trytond.i18n import gettext
 from trytond.model import (
-    ModelView, ModelSQL, Workflow, DeactivableMixin, fields, Unique)
+    DeactivableMixin, MatchMixin, ModelSQL, ModelView, Unique, Workflow,
+    fields, sequence_ordered)
 from trytond.model.exceptions import AccessError
-from trytond.pyson import Eval, Bool, Id
-from trytond.transaction import Transaction
-from trytond.pool import Pool
-from trytond.tools import reduce_ids, grouped_slice, lstrip_wildcard
-from trytond.tools.multivalue import migrate_property
 from trytond.modules.company.model import (
     CompanyMultiValueMixin, CompanyValueMixin)
+from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool
+from trytond.pyson import Bool, Eval, Id
+from trytond.tools import (
+    grouped_slice, is_full_text, lstrip_wildcard, reduce_ids)
+from trytond.transaction import Transaction
 
 STATES = {
-    'readonly': Eval('state') == 'close',
+    'readonly': Eval('state') == 'closed',
 }
-DEPENDS = ['state']
 
 
 class Journal(
-        DeactivableMixin, ModelSQL, ModelView, CompanyMultiValueMixin):
+        DeactivableMixin, MatchMixin,
+        sequence_ordered('matching_sequence', "Matching Sequence"),
+        ModelSQL, ModelView, CompanyMultiValueMixin):
     'Journal'
     __name__ = 'account.journal'
     name = fields.Char('Name', size=None, required=True, translate=True)
@@ -50,17 +52,18 @@ class Journal(
                 }))
     sequences = fields.One2Many(
         'account.journal.sequence', 'journal', "Sequences")
-    debit = fields.Function(fields.Numeric('Debit',
-            digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_debit_credit_balance')
-    credit = fields.Function(fields.Numeric('Credit',
-            digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_debit_credit_balance')
-    balance = fields.Function(fields.Numeric('Balance',
-            digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_debit_credit_balance')
-    currency_digits = fields.Function(fields.Integer('Currency Digits'),
-        'get_currency_digits')
+    debit = fields.Function(Monetary(
+            "Debit", currency='currency', digits='currency'),
+        'get_debit_credit_balance')
+    credit = fields.Function(Monetary(
+            "Credit", currency='currency', digits='currency'),
+        'get_debit_credit_balance')
+    balance = fields.Function(Monetary(
+            "Balance", currency='currency', digits='currency'),
+        'get_debit_credit_balance')
+
+    currency = fields.Function(fields.Many2One(
+            'currency.currency', "Currency"), 'get_currency')
 
     @classmethod
     def __setup__(cls):
@@ -69,33 +72,34 @@ class Journal(
 
     @classmethod
     def default_sequence(cls, **pattern):
-        return None
+        return cls.multivalue_model('sequence').default_sequence()
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        if clause[1].startswith('!') or clause[1].startswith('not '):
+        _, operator, operand, *extra = clause
+        if operator.startswith('!') or operator.startswith('not '):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
-        code_value = clause[2]
-        if clause[1].endswith('like'):
-            code_value = lstrip_wildcard(clause[2])
+        code_value = operand
+        if operator.endswith('like') and is_full_text(operand):
+            code_value = lstrip_wildcard(operand)
         return [bool_op,
-            ('code', clause[1], code_value) + tuple(clause[3:]),
-            (cls._rec_name,) + tuple(clause[1:]),
+            ('code', operator, code_value, *extra),
+            (cls._rec_name, operator, operand, *extra),
             ]
 
     @classmethod
-    def get_currency_digits(cls, journals, name):
+    def get_currency(cls, journals, name):
         pool = Pool()
         Company = pool.get('company.company')
         company_id = Transaction().context.get('company')
         if company_id:
             company = Company(company_id)
-            digits = company.currency.digits
+            currency_id = company.currency.id
         else:
-            digits = 2
-        return dict.fromkeys([j.id for j in journals], digits)
+            currency_id = None
+        return dict.fromkeys([j.id for j in journals], currency_id)
 
     @classmethod
     def get_debit_credit_balance(cls, journals, names):
@@ -150,16 +154,46 @@ class Journal(
                     debit - credit)
         return result
 
+    @classmethod
+    def find(cls, pattern):
+        for journal in cls.search(
+                [],
+                order=[
+                    ('matching_sequence', 'ASC'),
+                    ('id', 'ASC'),
+                    ]):
+            if journal.match(pattern):
+                return journal
+
+    @classmethod
+    def write(cls, *args):
+        pool = Pool()
+        Move = pool.get('account.move')
+        actions = iter(args)
+        for journals, values in zip(actions, actions):
+            if 'type' in values:
+                for sub_journals in grouped_slice(journals):
+                    moves = Move.search([
+                            ('journal', 'in', [j.id for j in sub_journals]),
+                            ('state', '=', 'posted')
+                            ], order=[], limit=1)
+                    if moves:
+                        move, = moves
+                        raise AccessError(gettext(
+                                'account.msg_journal_account_moves',
+                                journal=move.journal.rec_name))
+        super().write(*args)
+
 
 class JournalSequence(ModelSQL, CompanyValueMixin):
     "Journal Sequence"
     __name__ = 'account.journal.sequence'
     journal = fields.Many2One(
-        'account.journal', "Journal", ondelete='CASCADE', select=True,
+        'account.journal', "Journal", ondelete='CASCADE',
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
     sequence = fields.Many2One(
         'ir.sequence', "Sequence",
         domain=[
@@ -167,25 +201,16 @@ class JournalSequence(ModelSQL, CompanyValueMixin):
                 Id('account', 'sequence_type_account_journal')),
             ('company', 'in', [Eval('company', -1), None]),
             ],
-        depends=['company'])
+        depends={'company'})
 
     @classmethod
-    def __register__(cls, module_name):
-        exist = backend.TableHandler.table_exist(cls._table)
-
-        super(JournalSequence, cls).__register__(module_name)
-
-        if not exist:
-            cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('sequence')
-        value_names.append('sequence')
-        fields.append('company')
-        migrate_property(
-            'account.journal', field_names, cls, value_names,
-            parent='journal', fields=fields)
+    def default_sequence(cls):
+        pool = Pool()
+        ModelData = pool.get('ir.model.data')
+        try:
+            return ModelData.get_id('account', 'sequence_account_journal')
+        except KeyError:
+            return None
 
 
 class JournalCashContext(ModelView):
@@ -205,14 +230,14 @@ class JournalPeriod(
     'Journal - Period'
     __name__ = 'account.journal.period'
     journal = fields.Many2One('account.journal', 'Journal', required=True,
-            ondelete='CASCADE', states=STATES, depends=DEPENDS)
+            ondelete='CASCADE', states=STATES)
     period = fields.Many2One('account.period', 'Period', required=True,
-            ondelete='CASCADE', states=STATES, depends=DEPENDS)
+            ondelete='CASCADE', states=STATES)
     icon = fields.Function(fields.Char('Icon'), 'get_icon')
     state = fields.Selection([
         ('open', 'Open'),
-        ('close', 'Close'),
-        ], 'State', readonly=True, required=True)
+        ('closed', 'Closed'),
+        ], 'State', readonly=True, required=True, sort=False)
 
     @classmethod
     def __setup__(cls):
@@ -223,8 +248,8 @@ class JournalPeriod(
                 'account.msg_journal_period_unique'),
             ]
         cls._transitions |= set((
-                ('open', 'close'),
-                ('close', 'open'),
+                ('open', 'closed'),
+                ('closed', 'open'),
                 ))
         cls._buttons.update({
                 'close': {
@@ -232,20 +257,20 @@ class JournalPeriod(
                     'depends': ['state'],
                     },
                 'reopen': {
-                    'invisible': Eval('state') != 'close',
+                    'invisible': Eval('state') != 'closed',
                     'depends': ['state'],
                     },
                 })
         cls.active.states = STATES
-        cls.active.depends = DEPENDS
 
     @classmethod
-    def __register__(cls, module_name):
-        super(JournalPeriod, cls).__register__(module_name)
-
-        table = cls.__table_handler__(cls, module_name)
-        # Migration from 4.2: remove name column
-        table.drop_column('name')
+    def __register__(cls, module):
+        cursor = Transaction().connection.cursor()
+        t = cls.__table__()
+        super().__register__(module)
+        # Migration from 6.8: rename state close to closed
+        cursor.execute(
+            *t.update([t.state], ['closed'], where=t.state == 'close'))
 
     @staticmethod
     def default_state():
@@ -268,7 +293,7 @@ class JournalPeriod(
     def get_icon(self, name):
         return {
             'open': 'tryton-account-open',
-            'close': 'tryton-account-close',
+            'closed': 'tryton-account-close',
             }.get(self.state)
 
     @classmethod
@@ -301,7 +326,7 @@ class JournalPeriod(
     def write(cls, *args):
         actions = iter(args)
         for journal_periods, values in zip(actions, actions):
-            if (values != {'state': 'close'}
+            if (values != {'state': 'closed'}
                     and values != {'state': 'open'}):
                 cls._check(journal_periods)
             if values.get('state') == 'open':
@@ -321,7 +346,7 @@ class JournalPeriod(
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('close')
+    @Workflow.transition('closed')
     def close(cls, periods):
         '''
         Close journal - period
